from typing import Optional, List, Any

import numpy as np


def compute_atom_distance_matrix(coordinates: np.ndarray) -> np.ndarray:
    """Compute atom-to-atom distance for each conformer

    Parameters
    ----------
    coordinates: numpy.ndarray
        3D matrix of coordinates, with shape (n_conformers, n_atoms, 3)

    Returns
    -------
    distances: numpy.ndarray
        3D matrix of distances, with shape (n_conformers, n_atoms, n_atoms)
    """
    dist_sq = np.einsum("ijk,ilk->ijl", coordinates, coordinates)
    diag = np.einsum("ijj->ij", dist_sq)
    a, b = diag.shape
    dist_sq += dist_sq - diag.reshape((a, 1, b)) - diag.reshape((a, b, 1))
    diag[:] = -0.0
    return np.sqrt(-dist_sq)


def get_maximally_diverse_indices(
    distance_matrix: np.ndarray,
    distance_threshold: float = 0.05,
    n_indices: Optional[int] = None,
) -> List[int]:
    """Greedily select maximally diverse indices from distance_matrix

    Parameters
    ----------
    distance_matrix: numpy.ndarray
        2D square distance matrix with shape (n_items, n_items)
    distance_threshold: float
        If any item is below this threshold to any other item,
        they are considered too similar and only one will be included
        in the output
    n_indices: int
        Number of items to output

    Returns
    -------
    indices: List[int]
        List of indices of maximally diverse items.
    """
    n_distances = len(distance_matrix)
    if distance_matrix.shape != (n_distances, n_distances):
        raise ValueError("`distance_matrix` should be square distance matrix")

    if n_indices is None:
        n_indices = n_distances
    n_indices = min(n_indices, n_distances)

    selected_indices = [0]
    for i in range(n_indices - 1):
        selected_rms = distance_matrix[selected_indices]
        any_too_close = np.any(selected_rms < distance_threshold, axis=0)
        if np.all(any_too_close):
            break

        rmsdist = np.where(any_too_close, -np.inf, selected_rms.sum(axis=0))
        selected_indices.append(rmsdist.argmax())

    return selected_indices

def isiterable(obj: Any) -> bool:
    if hasattr(obj, "__next__"):
        return True
    if hasattr(obj, "__iter__"):
        return True
    try:
        len(obj)
    except TypeError:
        pass
    else:
        return True
    return False


from typing import Union, List, Tuple
from typing_extensions import Literal

from rdkit import Chem as rdChem
from rdkit.Chem import rdForceFieldHelpers as rdFF
from rdkit.ForceField.rdForceField import ForceField as RDForceField

RDKIT_FF_NAMES = ["UFF", "MMFF", "MMFF94", "MMFF94S"]
RDKIT_FF_NAME_TYPE = Literal[(*RDKIT_FF_NAMES,)]
RDKIT_FF_TYPE = Union[RDForceField, RDKIT_FF_NAME_TYPE]


def GetMoleculeForceField(
    mol: rdChem.Mol,
    forcefield: RDKIT_FF_NAME_TYPE = "MMFF94",
    ignoreInterfragInteractions: bool = True,
    vdwThresh: float = 10.0,
    nonBondedThresh: float = 100.0,
    confId: int = -1,
) -> RDForceField:
    """Get molecule force field with string

    Parameters
    ----------
    mol: rdkit.Chem.Mol
        Molecule of interest
    forcefield: Literal["UFF", "MMFF", "MMFF94", "MMFF94S"]
        Force field name. If an MMFF force field is used,
        you can choose the MMFF94 or MMFF94s variant.
        "MMFF" defaults to MMFF94.
    ignoreInterfragInteractions: bool
        If True, nonbonded terms between fragments will not be added to the forcefield.
    vdwThresh: float
        Used to exclude long-range van der Waals interactions.
        Only used for the UFF force field.
    nonBondedThresh: float
        Used to exclude long-range non-bonded interactions.
        Only used for the MMFF force field.
    confId: int
        Which conformer to optimize

    Returns
    -------
    rdkit.Chem.rdForceFieldHelpers.ForceField
    """
    if mol.GetNumConformers() == 0:
        raise ValueError("Molecule must have conformers to create a force field")

    forcefield = forcefield.upper()
    if forcefield not in RDKIT_FF_NAMES:
        raise ValueError(
            f"{forcefield} is not a supported force field. "
            "Supporte force fields: " + ", ".join(RDKIT_FF_NAMES)
        )
    if forcefield == "UFF":
        if not rdFF.UFFHasAllMoleculeParams(mol):
            raise ValueError(
                "MMFF force field does not contain all parameters "
                f"for given molecule: {rdChem.MolToSmarts(mol)}"
            )
        return rdFF.UFFGetMoleculeForceField(
            mol,
            vdwThresh=vdwThresh,
            confId=confId,
            ignoreInterfragInteractions=ignoreInterfragInteractions,
        )

    # associate MMFF with variants
    if forcefield == "MMFF":
        forcefield = "MMFF94"
    elif forcefield == "MMFF94S":
        forcefield = "MMFF94s"

    if not rdFF.MMFFHasAllMoleculeParams(mol):
        raise ValueError(
            "MMFF force field does not contain all parameters "
            f"for given molecule: {rdChem.MolToSmarts(mol)}"
        )
    molprops = rdFF.MMFFGetMoleculeProperties(mol, mmffVariant=forcefield)
    return rdFF.MMFFGetMoleculeForceField(
        mol,
        molprops,
        nonBondedThresh=nonBondedThresh,
        confId=confId,
        ignoreInterfragInteractions=ignoreInterfragInteractions,
    )


def FFOptimizeMolecule(
    mol: rdChem.Mol,
    forcefield: RDKIT_FF_TYPE = "MMFF94",
    maxIters: int = 200,
    vdwThresh: float = 10.0,
    nonBondedThresh: float = 100.0,
    confId: int = -1,
    ignoreInterfragInteractions: bool = True,
) -> int:
    """Optimize molecule

    Parameters
    ----------
    mol: rdkit.Chem.Mol
        Molecule of interest
    forcefield: Literal["UFF", "MMFF", "MMFF94", "MMFF94S"] or RDKit force field
        Force field name or RDKit force field.
        If an MMFF force field is used,
        you can choose the MMFF94 or MMFF94s variant.
        "MMFF" defaults to MMFF94.
    maxIters: int
        the maximum number of iterations
    ignoreInterfragInteractions: bool
        If True, nonbonded terms between fragments will not be added to the forcefield.
    vdwThresh: float
        Used to exclude long-range van der Waals interactions.
        Only used for the UFF force field.
    nonBondedThresh: float
        Used to exclude long-range non-bonded interactions.
        Only used for the MMFF force field.
    confId: int
        Which conformer to optimize

    Returns
    -------
    success: int
        0 if the optimization converged, 1 if more iterations are required.
    """
    if isinstance(forcefield, str):
        forcefield = GetMoleculeForceField(
            mol,
            forcefield=forcefield,
            vdwThresh=vdwThresh,
            nonBondedThresh=nonBondedThresh,
            confId=confId,
            ignoreInterfragInteractions=ignoreInterfragInteractions,
        )

    return rdFF.OptimizeMolecule(forcefield, maxIters=maxIters)


def FFOptimizeMoleculeConfs(
    mol: rdChem.Mol,
    forcefield: RDKIT_FF_TYPE = "MMFF94",
    maxIters: int = 200,
    numThreads: int = 1,
    vdwThresh: float = 10.0,
    nonBondedThresh: float = 100.0,
    confId: int = -1,
    ignoreInterfragInteractions: bool = True,
) -> List[Tuple[int, float]]:
    """Optimize molecule conformers

    Parameters
    ----------
    mol: rdkit.Chem.Mol
        Molecule of interest
    forcefield: Literal["UFF", "MMFF", "MMFF94", "MMFF94S"] or RDKit force field
        Force field name or RDKit force field.
        If an MMFF force field is used,
        you can choose the MMFF94 or MMFF94s variant.
        "MMFF" defaults to MMFF94.
    maxIters: int
        the maximum number of iterations
    numThreads: int
        the number of threads to use, only has an effect if the RDKit
        was built with thread support (defaults to 1).
        If set to zero, the max supported by the system will be used.
    ignoreInterfragInteractions: bool
        If True, nonbonded terms between fragments will not be added to the forcefield.
    vdwThresh: float
        Used to exclude long-range van der Waals interactions.
        Only used for the UFF force field.
    nonBondedThresh: float
        Used to exclude long-range non-bonded interactions.
        Only used for the MMFF force field.
    confId: int
        Which conformer to optimize

    Returns
    -------
    success_energies: List[Tuple[int, float]]
        List of tuples of (success, energy) for each conformer.
        The success is 0 if the optimization converged,
        1 if more iterations are required.
    """
    if isinstance(forcefield, str):
        forcefield = GetMoleculeForceField(
            mol,
            forcefield=forcefield,
            vdwThresh=vdwThresh,
            nonBondedThresh=nonBondedThresh,
            confId=confId,
            ignoreInterfragInteractions=ignoreInterfragInteractions,
        )

    return rdFF.OptimizeMoleculeConfs(
        mol,
        forcefield,
        maxIters=maxIters,
        numThreads=numThreads,
    )



"""
This is a bit of a hacky way to override RDKit functions
"""
import os
from typing import Optional

from rdkit.Chem.rdmolfiles import *  # type: ignore
from rdkit_utilities._io._rdmolfiles import *

CONTENT_PARSERS = {
    "fasta": MolFromFASTA,
    "helm": MolFromHELM,
    "mol": MolFromMolBlock,
    "mol2": MolFromMol2Block,
    "pdb": MolFromPDBBlock,
    "png": MolFromPNGString,
    "svg": MolFromRDKitSVG,
    "sequence": MolFromSequence,
    "seq": MolFromSequence,
    "smarts": MolFromSmarts,
    "smiles": MolFromSmiles,
    "tpl": MolFromTPLBlock,
}

FILE_PARSERS = {
    "mol": MolFromMolFile,
    "mol2": MolFromMol2File,
    "pdb": MolFromPDBFile,
    "png": MolFromPNGFile,
    "tpl": MolFromTPLFile,
}

ALL_RDKIT_PARSERS = sorted(set(CONTENT_PARSERS) | set(FILE_PARSERS))


def molecule_from_input(mol_input, *args, mol_format: Optional[str] = None, **kwargs):
    if os.path.isfile(mol_input):
        content_reader = FILE_PARSERS
        if not mol_format:
            mol_format = os.path.splitext(mol_input)[1][1:]
    else:
        content_reader = CONTENT_PARSERS
    if mol_format:
        try:
            reader = content_reader[mol_format.lower()]
        except KeyError:
            raise TypeError(
                f"Molecule format {mol_format} not supported. "
                f"Supported formats: " + ", ".join(ALL_RDKIT_PARSERS)
            )
        return reader(mol_input, *args, **kwargs)

    for reader in content_reader.values():
        try:
            mol = reader(mol_input, *args, **kwargs)
        except RuntimeError:
            pass
        else:
            if mol is not None:
                return mol

    raise TypeError(
        f"Could not create an RDKit molecule from {mol_input}. "
        "Try passing in a `mol_format`. "
        f"Supported formats: " + ", ".join(ALL_RDKIT_PARSERS)
    )



import pytest

from numpy.testing import assert_allclose

from rdkit_utilities import rdForceFieldHelpers


@pytest.mark.parametrize(
    "ff_name, energy",
    [
        ("UFF", 192350.97617245736),
        ("uff", 192350.97617245736),
        ("mmff", 69616788.17478018),
        ("MMFF94", 69616788.17478018),
        ("MMFF94S", 69616788.17478018),
    ],
)
def test_GetMoleculeForceField(methane, ff_name, energy):
    ff = rdForceFieldHelpers.GetMoleculeForceField(methane, forcefield=ff_name)
    assert_allclose(ff.CalcEnergy(), energy)


@pytest.mark.parametrize("ff", ["uff", "MMFF", "mmff94s"])
def test_FFOptimizeMolecule(ethane, ff):
    success = rdForceFieldHelpers.FFOptimizeMolecule(ethane, ff)
    assert success == 0


@pytest.mark.parametrize("ff", ["uff", "MMFF", "mmff94s"])
def test_FFOptimizeMoleculeConfs(ethane, ff):
    success_energies = rdForceFieldHelpers.FFOptimizeMoleculeConfs(ethane, ff)
    for conf in success_energies:
        assert conf[0] == 0



import pytest
import numpy as np

from rdkit import Chem as rdChem
from rdkit_utilities import rdmolops, rdmolfiles, rdchem


@pytest.fixture
def propylparaben():
    rdmol = rdmolfiles.MolFromSmiles(
        (
            "[H:1]-[O:2]-[c:3]1:[c:4](-[H:5]):[c:6](-[H:7])"
            ":[c:8](-[C:13](=[O:14])-[O:15]-[C:16](-[H:17])"
            "(-[H:18])-[C:19](-[H:20])(-[H:21])-[C:22](-[H:23])"
            "(-[H:24])-[H:25]):[c:9](-[H:10]):[c:11]:1-[H:12]"
        ),
        removeHs=False,
    )
    rdmol = rdmolops.OrderByMapNumber(rdmol, clearAtomMapNumbers=True)
    return rdmol


@pytest.fixture
def methane():
    mol = rdmolfiles.MolFromSmarts(
        "[H:2]-[C:1](-[H:3])(-[H:4])-[H:5]",
        orderByMapNumber=True,
    )
    for i in range(5):
        conf = np.ones((5, 3), dtype=float) * np.arange(5).reshape((5, 1))
        conf[i] += 10
        rdchem.AddConformerWithCoordinates(mol, conf)
    rdChem.SanitizeMol(mol)
    return mol


@pytest.fixture()
def ethane():
    from rdkit.Chem import AllChem

    mol = rdmolfiles.MolFromSmiles("CC")
    mol = rdChem.AddHs(mol)
    rdChem.SanitizeMol(mol)
    AllChem.EmbedMolecule(mol)
    return mol



"""
Functions to do with molecules that are analogous to rdkit.Chem.rdchem.
"""

from typing import List, Tuple, Union, Dict, Any

from rdkit import Chem as rdChem
import numpy as np

from .utilities import isiterable


def AddConformerWithCoordinates(
    mol: rdChem.Mol,
    coordinates: np.ndarray,
) -> int:
    """Add conformer to molecule with coordinates in angstrom"""
    from rdkit import Geometry

    n_atoms = mol.GetNumAtoms()
    coord_type = np.asarray(coordinates)
    if coord_type.shape != (n_atoms, 3):
        raise ValueError(
            f"Shape of coordinates must be ({n_atoms}, 3). "
            f"Given array with shape {coord_type.shape}"
        )
    conformer = rdChem.Conformer(n_atoms)
    for i, xyz in enumerate(coordinates):
        x, y, z = map(float, xyz)
        conformer.SetAtomPosition(i, Geometry.Point3D(x, y, z))
    conformer.SetId(mol.GetNumConformers())
    return mol.AddConformer(conformer)


def GetSymmetricAtomIndices(
    mol: rdChem.Mol,
    maxMatches: int = 10000,
) -> List[Tuple[int, ...]]:
    """Get atom indices of symmetric atoms

    Returns
    -------
    symmetric_indices: List[Tuple[int, ...]]
        In this list, one item is a sorted tuple of indices,
        where each index indicates an atom that is symmetric
        to all the other indices in the tuple.
    """
    # take care of resonance
    matches = [
        match
        for resMol in rdChem.ResonanceMolSupplier(mol)
        for match in resMol.GetSubstructMatches(
            mol, uniquify=False, maxMatches=maxMatches
        )
    ]
    atom_symmetries = set(tuple(sorted(set(x))) for x in zip(*matches))
    return sorted([x for x in atom_symmetries if len(x) > 1])


def GetTaggedSubstructMatches(
    mol: rdChem.Mol,
    query: rdChem.Mol,
    uniquify: bool = False,
    useChirality: bool = False,
    useQueryQueryMatches: bool = False,
    maxMatches: int = 1000,
    mappingAsTaggedDict: bool = False,
) -> Union[List[Tuple[int, ...]], List[Dict[int, int]]]:
    """Only return tagged atoms in substruct match, in tag order"""
    matches = mol.GetSubstructMatches(
        query,
        uniquify=uniquify,
        useChirality=useChirality,
        useQueryQueryMatches=useQueryQueryMatches,
        maxMatches=maxMatches,
    )
    map_numbers = np.array([a.GetAtomMapNum() for a in query.GetAtoms()])
    not_zero = np.where(map_numbers)[0]
    matches = [
        tuple(x for i, x in enumerate(match) if i in not_zero) for match in matches
    ]
    matches = sorted(set(matches))
    map_numbers = map_numbers[not_zero]
    order = np.argsort(map_numbers)
    matches = [tuple(match[i] for i in order) for match in matches]
    if mappingAsTaggedDict:
        map_numbers = list(map(int, map_numbers[order]))
        return [dict(zip(map_numbers, match)) for match in matches]
    return matches


def SetPropsFromDict(obj: Union[rdChem.Mol, rdChem.Atom], properties: Dict[str, Any]):
    """Set properties from dict, analogous to GetPropsAsDict()"""
    for key, val in properties.items():
        if key == "__computedProps":
            continue
        valtype = type(val)
        if val is False or val is True or np.issubdtype(valtype, np.bool_):
            obj.SetBoolProp(key, bool(val))
        # exhaustively list np types for Windows
        elif (isinstance(val, (int, np.int_, np.int16, np.int32, np.int64, np.int0, np.intp, np.intc))
              or np.issubdtype(valtype, np.int_)):
            obj.SetIntProp(key, int(val))
        elif (isinstance(val, (float, np.float_, np.float16, np.float32, np.float64))
              or np.issubdtype(valtype, np.float_)):
            obj.SetDoubleProp(key, float(val))
        elif isinstance(val, str) or np.issubdtype(valtype, str):
            obj.SetProp(key, val)
        elif isinstance(val, rdChem.rdBase._vecti):
            obj.SetExplicitBitVectProp(key, val)
        else:
            raise ValueError(f"No setter function for {val} of type {valtype}")



import itertools
from typing import Optional, Dict, List, Union, Set, Tuple
from typing_extensions import Literal

import numpy as np

from rdkit import Chem as rdChem
from rdkit.Chem import AllChem as rdAllChem

from rdkit_utilities import rdForceFieldHelpers


def GetExclusions(mol: rdChem.Mol) -> Set[Tuple[int, int]]:
    """Get exclusions with a molecule between atoms 2- and 3- sites away"""
    exclusions = set()
    for i, atom in enumerate(mol.GetAtoms()):
        partners = [b.GetOtherAtomIdx(i) for b in atom.GetBonds()]
        exclusions |= set(tuple(sorted([i, x])) for x in partners)
        exclusions |= set(itertools.combinations(sorted(partners), 2))
    return exclusions


def GenerateConformers(
    mol: rdChem.Mol,
    numConfs: int = 10,
    numConfPool: Optional[int] = None,
    maximizeDiversity: bool = False,
    selectELFConfs: bool = False,
    ELFPercentage: float = 2.0,
    pruneRmsThresh: float = -1.0,
    diverseRmsThresh: float = 0.05,
    energyWindow: float = -1.0,
    maxAttempts: int = 0,
    randomSeed: int = -1,
    clearConfs: bool = True,
    useRandomCoords: bool = False,
    boxSizeMult: float = 2.0,
    randNegEig: bool = True,
    numZeroFail: int = 1,
    coordMap: Dict[int, Union[np.ndarray, List[float]]] = {},
    forceTol: float = 0.001,
    ignoreSmoothingFailures: bool = False,
    enforceChirality: bool = True,
    numThreads: int = 1,
    useExpTorsionAnglePrefs: bool = True,
    useBasicKnowledge: bool = True,
    printExpTorsionAngles: bool = False,
    useSmallRingTorsions: bool = False,
    useMacrocycleTorsions: bool = False,
    ETversion: int = 1,
    optimizeConfs: bool = False,
    forcefield: rdForceFieldHelpers.RDKIT_FF_TYPE = "MMFF94",
    optMaxIters: int = 200,
    ffVdwThresh: float = 10.0,
    ffNonBondedThresh: float = 100,
    ffIgnoreInterfragInteractions: bool = True,
    removeTransAcidConformers: bool = True,
) -> List[int]:
    """
    Generate conformers on molecule in-place.

    Parameters
    ----------
    numConfs: int
        Max number of conformers to end up with
    numConfPool: Optional[int]
        Max number of conformers to initially generate,
        for ELF or RMS selection.
    maximizeDiversity: bool
        Whether to maximize diversity via heavy-atom RMS
    selectELFConfs: bool
        Whether to select conformers using the electrostatically
        least-interacting functional group (ELF) technique
    ELFPercentage: float
        % (out of 100) of conformers to keep from `numConfPool`
        when selecting ELF conformers
    pruneRmsThresh: float
        RMS threshold to keep conformers initially generated
        from `numConfPool`
    diverseRmsThresh: float
        Minimum RMS threshold for distinguishing conformers
        when selecting maximally diverse conformers
    energyWindow: float
        If positive, only conformers within this energy
        window as determined by `forcefield` will be kept.
    optimizeConfs: bool
        Whether to optimize conformers generated initially,
        prior to selecting ELF and RMS diverse conformers.
    forcefield: Literal["UFF", "MMFF94", "MMFF94S"]
        Which force field to use for optimizing conformers
        and calculating energies for `energyWindow`.
    optMaxIters: int
        Maximum iterations for optimization
    ffVdwThresh: float
        VdW threshold for optimization and calculating energies
        using force field
    ffNonBondedThresh: float
        Non-bonded threshold for optimization and calculating energies
        using force field
    ffIgnoreInterfragInteractions: bool
        If True, nonbonded terms between fragments will not be added
        to the forcefield.
    **kwargs:
        See the RDKit documentation for
        ``rdkit.Chem.rdDistGeom.EmbedMultipleConfs``
        for more.
    """
    from .rdmolops import KeepConformerIds

    if not numConfPool:
        numConfPool = numConfs
        if maximizeDiversity or energyWindow > 0:
            numConfPool *= 1000

    coordMap = {k: list(map(float, v)) for k, v in coordMap.items()}

    rdAllChem.EmbedMultipleConfs(
        mol,
        numConfs=numConfPool,
        maxAttempts=maxAttempts,
        randomSeed=randomSeed,
        clearConfs=clearConfs,
        useRandomCoords=useRandomCoords,
        boxSizeMult=boxSizeMult,
        randNegEig=randNegEig,
        numZeroFail=numZeroFail,
        pruneRmsThresh=pruneRmsThresh,
        coordMap=coordMap,
        forceTol=forceTol,
        ignoreSmoothingFailures=ignoreSmoothingFailures,
        enforceChirality=enforceChirality,
        numThreads=numThreads,
        useExpTorsionAnglePrefs=useExpTorsionAnglePrefs,
        useBasicKnowledge=useBasicKnowledge,
        printExpTorsionAngles=printExpTorsionAngles,
        useSmallRingTorsions=useSmallRingTorsions,
        useMacrocycleTorsions=useMacrocycleTorsions,
        ETversion=ETversion,
    )

    if optimizeConfs:
        rdForceFieldHelpers.FFOptimizeMoleculeConfs(
            mol,
            forcefield=forcefield,
            numThreads=numThreads,
            maxIters=optMaxIters,
            vdwThresh=ffVdwThresh,
            nonBondedThresh=ffNonBondedThresh,
            ignoreInterfragInteractions=ffIgnoreInterfragInteractions,
        )

    if energyWindow > 0:
        RemoveConformersOutsideEnergyWindow(
            mol,
            energyWindow=energyWindow,
            forcefield=forcefield,
            numThreads=numThreads,
            ffVdwThresh=ffVdwThresh,
            ffNonBondedThresh=ffNonBondedThresh,
            ffIgnoreInterfragInteractions=ffIgnoreInterfragInteractions,
        )

    if selectELFConfs:
        SelectELFConformers(
            mol,
            ELFPercentage=ELFPercentage,
            removeTransAcidConformers=removeTransAcidConformers,
        )

    if maximizeDiversity:
        SelectDiverseConformers(
            mol, numConfs=numConfs, diverseRmsThresh=diverseRmsThresh
        )

    conformer_ids = [conf.GetId() for conf in mol.GetConformers()]
    KeepConformerIds(mol, conformer_ids[:numConfs])
    return mol.GetNumConformers()


def RemoveTransAcidConformers(mol: rdChem.Mol):
    """Remove conformers from molecule if they contain a trans carboxylic acid"""
    from rdkit.Chem.rdMolTransforms import GetDihedralRad  # type: ignore[import]
    from rdkit_utilities import rdmolfiles
    from .rdmolops import KeepConformerIds

    query = rdmolfiles.MolFromSmarts(
        "[#6X3:2](=[#8:1])(-[#8X2H1:3]-[#1:4])",
        orderByMapNumber=True,
    )
    matches = mol.GetSubstructMatches(query, uniquify=False)
    to_keep = []
    for conformer in mol.GetConformers():
        for match in matches:
            angle = GetDihedralRad(conformer, *match)
            if angle > np.pi / 2.0:
                break
        else:
            to_keep.append(conformer.GetId())
    KeepConformerIds(mol, to_keep)


def SelectELFConformers(
    mol: rdChem.Mol, ELFPercentage: float = 2.0, removeTransAcidConformers: bool = True
):
    """Keep percentage of conformers, ranked by electrostatic energy, in-place.

    The other conformers are deleted.
    """
    from .rdmolops import KeepConformerIds

    if removeTransAcidConformers:
        RemoveTransAcidConformers(mol)

    SortConformersByElectrostaticInteraction(mol)
    goal = int(mol.GetNumConformers() * ELFPercentage / 100)
    n_keep = max(1, goal)
    conformer_ids = [c.GetId() for c in mol.GetConformers()]
    keep_ids = conformer_ids[:n_keep]
    KeepConformerIds(mol, keep_ids)


def SelectDiverseConformers(
    mol: rdChem.Mol,
    numConfs: int = 10,
    diverseRmsThresh: float = 0.05,
):
    """Select conformers for max diversity by RMS, in-place.

    This uses a greedy algorithm. Unselected conformers are deleted.
    """
    from .rdMolAlign import GetBestConformerRMS
    from .rdmolops import KeepConformerIds
    from .utilities import get_maximally_diverse_indices

    rms_matrix = GetBestConformerRMS(mol, heavyAtomsOnly=True)
    diverse_indices = get_maximally_diverse_indices(
        rms_matrix, distance_threshold=diverseRmsThresh, n_indices=numConfs
    )
    conformer_ids = [conf.GetId() for conf in mol.GetConformers()]
    keep_ids = [conformer_ids[i] for i in diverse_indices]
    KeepConformerIds(mol, keep_ids)
    for i, conf in enumerate(mol.GetConformers()):
        conf.SetId(i)


def SelectDiverseELFConformers(
    mol: rdChem.Mol,
    numConfs: int = 10,
    ELFPercentage: float = 2.0,
    diverseRmsThresh: float = 0.05,
    removeTransAcidConformers: bool = True,
):
    """Select diverse conformers using the ELF technique, in-place.

    Other conformers are deleted from the molecule.
    """
    SelectELFConformers(
        mol,
        ELFPercentage=ELFPercentage,
        removeTransAcidConformers=removeTransAcidConformers,
    )
    SelectDiverseConformers(mol, numConfs=numConfs, diverseRmsThresh=diverseRmsThresh)


def SortConformersByEnergy(
    mol: rdChem.Mol,
    forcefield: rdForceFieldHelpers.RDKIT_FF_TYPE = "MMFF94",
    numThreads: int = 1,
    ffVdwThresh: float = 10.0,
    ffNonBondedThresh: float = 100,
    ffIgnoreInterfragInteractions: bool = True,
    reverse: bool = False,
) -> List[float]:
    """Sort conformers in-place on molecule by force field energy.

    Returns
    -------
    energies: List[float]

    Raises
    ------
    ValueError
        If the force field cannot calculate energies
    """
    from .rdmolops import ReorderConformers

    result = rdForceFieldHelpers.FFOptimizeMoleculeConfs(
        mol,
        forcefield=forcefield,
        numThreads=numThreads,
        maxIters=0,
        vdwThresh=ffVdwThresh,
        nonBondedThresh=ffNonBondedThresh,
        ignoreInterfragInteractions=ffIgnoreInterfragInteractions,
    )
    energies = [x[1] for x in result]
    if any(x == -1 for x in energies):
        raise ValueError(
            f"Could not use {forcefield} to calculate energies "
            f"for {rdChem.MolToSmarts(mol)}. "
            "Cannot sort conformers by energy."
        )
    order = np.argsort(energies)
    if reverse:
        order = order[::-1]
    ReorderConformers(mol, order)
    return [energies[i] for i in order]


def CalculateMMFFCharges(
    mol: rdChem.Mol,
    forcefield: Literal["MMFF94", "MMFF94s"] = "MMFF94",
    normalize_partial_charges: bool = True,
) -> np.ndarray:
    """Calculate MMFF charges for molecule"""
    forcefield = forcefield.upper()
    if forcefield == "MMFF94S":
        forcefield = "MMFF94s"
    mps = rdAllChem.MMFFGetMoleculeProperties(mol, forcefield)
    if mps is None:
        molstr = rdChem.MolToSmiles(mol)
        raise ValueError(f"MMFF charges could not be computed for {molstr}")
    n_atoms = mol.GetNumAtoms()
    charges = np.array([mps.GetMMFFPartialCharge(i) for i in range(n_atoms)])

    if normalize_partial_charges:
        total_charge = rdChem.GetFormalCharge(mol)
        partial_charges = charges.sum()
        offset = (total_charge - partial_charges) / n_atoms
        charges += offset
    return charges


def CalculateElectrostaticEnergy(
    mol: rdChem.Mol,
    forcefield: Literal["MMFF94", "MMFF94s"] = "MMFF94",
) -> np.ndarray:
    """Calculate electrostatic energy of all conformers using force field

    Returns
    -------
    energies: numpy.ndarray
        Array of energies, one for each conformer
    """
    from .utilities import compute_atom_distance_matrix

    conformers = np.array([c.GetPositions() for c in mol.GetConformers()])
    distances = compute_atom_distance_matrix(conformers)
    inverse_distances = np.reciprocal(
        distances, out=np.zeros_like(distances), where=~np.isclose(distances, 0)
    )

    charges = np.abs(CalculateMMFFCharges(mol, forcefield)).reshape(-1, 1)
    charge_products = charges @ charges.T

    excl_i, excl_j = zip(*GetExclusions(mol))
    charge_products[(excl_i, excl_j)] = 0.0
    charge_products[(excl_j, excl_i)] = 0.0

    energies = inverse_distances * charge_products
    return 0.5 * energies.sum(axis=(1, 2))


def SortConformersByElectrostaticInteraction(
    mol: rdChem.Mol,
    forcefield: Literal["MMFF94", "MMFF94s"] = "MMFF94",
):
    """Sort conformers in-place by electrostatic energy using MMFF"""
    from .rdmolops import ReorderConformers

    energies = CalculateElectrostaticEnergy(mol, forcefield)
    order = np.argsort(energies)
    ReorderConformers(mol, order)


def RemoveConformersOutsideEnergyWindow(
    mol: rdChem.Mol,
    energyWindow: float = 30.0,
    forcefield: rdForceFieldHelpers.RDKIT_FF_TYPE = "MMFF94",
    numThreads: int = 1,
    ffVdwThresh: float = 10.0,
    ffNonBondedThresh: float = 100,
    ffIgnoreInterfragInteractions: bool = True,
) -> int:
    """Remove conformers outside energy window using force field, in-place

    The window covers an energy range from the conformer with the lowest
    energy. The molecule also has its conformers sorted by force field
    energy.

    Returns
    -------
    numConfs: int
        The number of remaining conformers
    """
    from .rdmolops import KeepConformerIds

    if energyWindow < 0:
        return mol.GetNumConformers()

    energies = SortConformersByEnergy(
        mol,
        forcefield=forcefield,
        numThreads=numThreads,
        ffVdwThresh=ffVdwThresh,
        ffNonBondedThresh=ffNonBondedThresh,
        ffIgnoreInterfragInteractions=ffIgnoreInterfragInteractions,
        reverse=False,
    )
    lowest = energies[0]
    highest = lowest + energyWindow
    conf_ids = [conf.GetId() for conf in mol.GetConformers()]
    mask = np.array(energies) <= highest
    keep = np.array(conf_ids)[mask]
    KeepConformerIds(mol, keep)

    return mol.GetNumConformers()



import pytest
from rdkit_utilities import rdmolfiles
from rdkit_utilities.tests.datafiles import CCH_PDB


@pytest.mark.parametrize(
    "smiles, n_heavy_atoms, n_all_atoms",
    [
        ("[C:3][C:2][H:1]", 2, 3),
    ],
)
def test_MolFromSmiles(smiles, n_heavy_atoms, n_all_atoms):
    heavy_mol = rdmolfiles.MolFromSmiles(smiles)
    assert heavy_mol.GetNumAtoms() == n_heavy_atoms
    all_mol = rdmolfiles.MolFromSmiles(smiles, removeHs=False)
    assert all_mol.GetNumAtoms() == n_all_atoms


@pytest.mark.parametrize(
    "orderByMapNumber, clearAtomMapNumbers, firstElement, firstNum",
    [
        (False, False, "C", 3),
        (False, True, "C", 0),
        (True, False, "H", 1),
        (True, True, "H", 0),
    ],
)
def test_MolFromSmarts(orderByMapNumber, clearAtomMapNumbers, firstElement, firstNum):
    rdmol = rdmolfiles.MolFromSmarts(
        "[C:3][C:2][H:1]",
        orderByMapNumber=orderByMapNumber,
        clearAtomMapNumbers=clearAtomMapNumbers,
    )
    first = rdmol.GetAtomWithIdx(0)
    assert first.GetSymbol() == firstElement
    assert first.GetAtomMapNum() == firstNum


@pytest.mark.parametrize(
    "mol_format, mol_input, order_atoms, map_number",
    [
        ("smiles", "[C:3][C:2][H:1]", True, 2),
        ("smiles", "[C:3][C:2][H:1]", False, 3),
        ("smarts", "[C:3][C:2][H:1]", True, 1),
        ("smarts", "[C:3][C:2][H:1]", False, 3),
        (None, "[C:3][C:2][H:1]", True, 1),
        (None, "[C:3][C:2][H:1]", False, 3),
        ("pdb", CCH_PDB, True, 0),
        ("pdb", CCH_PDB, False, 0),
        (None, CCH_PDB, True, 0),
        (None, CCH_PDB, False, 0),
    ],
)
def test_mol_from_input(mol_format, mol_input, order_atoms, map_number):
    rdmol = rdmolfiles.MolFromInput(
        mol_input, inputFormat=mol_format, orderByMapNumber=order_atoms
    )
    first = rdmol.GetAtomWithIdx(0)
    assert first.GetAtomMapNum() == map_number



from typing import Tuple, List
import itertools

from rdkit import Chem as rdChem
from rdkit.Chem import rdChemReactions


def ClickReaction(
    reaction: rdChemReactions.ChemicalReaction,
    reactants: Tuple[rdChem.Mol, ...],
) -> List[rdChem.Mol]:
    """Run a "click" reaction by manually building a molecule

    Occasionally an RDKit ChemicalReaction.RunReactants won't work.
    If a product is fully tagged and specified, you can try running
    this click reaction instead.

    Note that this does not do any fancy chemistry inferring.
    You will need to specify any hydrogens you want to remove, for example.
    """
    rdproducts = reaction.GetProducts()
    if len(rdproducts) != 1:
        raise ValueError(
            "This function is only supported for reactions with 1 product. "
            f"The reaction given has {len(rdproducts)} products"
        )

    rdproduct = rdproducts[0]
    product_map_numbers = [atom.GetAtomMapNum() for atom in rdproduct.GetAtoms()]
    if 0 in product_map_numbers:
        raise ValueError(
            "Reaction product must have every atom tagged with a map number. "
            f"Given: {rdChem.MolToSmarts(rdproduct)}"
        )

    rdreactants = list(reaction.GetReactants())
    if not len(rdreactants) == len(reactants):
        raise ValueError(
            "`reactants` must have same length as reactants in `reaction`. "
            f"Given {len(reactants)} reactants for {len(rdreactants)} in reaction"
        )

    ORIGINAL_MAP_NUMBER = "_original_map_num"

    edited_reactants = []
    for rdreactant, reactant in zip(rdreactants, reactants):
        rd_map_numbers = [atom.GetAtomMapNum() for atom in rdreactant.GetAtoms()]
        indices_to_keep = []
        indices_to_delete = []
        for i, x in enumerate(rd_map_numbers):
            if x not in product_map_numbers:
                indices_to_delete.append(i)
            else:
                indices_to_keep.append(i)

        reactant_list = []
        for match in reactant.GetSubstructMatches(rdreactant):
            # cut bonds to atoms we want to delete
            to_delete = sorted([match[i] for i in indices_to_delete])
            reactant_ = rdChem.RWMol(reactant)
            for atom in reactant_.GetAtoms():
                atom.SetIntProp(ORIGINAL_MAP_NUMBER, atom.GetAtomMapNum())
                atom.SetAtomMapNum(0)

            for i in indices_to_keep:
                atom = reactant_.GetAtomWithIdx(match[i])
                atom.SetAtomMapNum(rd_map_numbers[i])

            for i in to_delete[::-1]:
                atom_ = reactant_.GetAtomWithIdx(i)
                for bond in atom_.GetBonds():
                    reactant_.RemoveBond(bond.GetBeginAtomIdx(), bond.GetEndAtomIdx())
            # keep only one fragment
            mapping = []
            fragments = rdChem.GetMolFrags(
                reactant_, asMols=True, sanitizeFrags=False, fragsMolAtomMapping=mapping
            )
            for i, fragment_indices in enumerate(mapping):
                if all(match[x] in fragment_indices for x in indices_to_keep):
                    reactant_list.append(fragments[i])
                    break
        edited_reactants.append(reactant_list)

    rwproducts: List[rdChem.Mol] = []
    for group in itertools.product(*edited_reactants):
        rwproduct = rdChem.Mol()
        for reactant in group:
            rwproduct = rdChem.CombineMols(rwproduct, reactant)
            rwproduct = rdChem.RWMol(rwproduct)

        index_to_map_numbers = {
            atom.GetIdx(): atom.GetAtomMapNum()
            for atom in rwproduct.GetAtoms()
            if atom.GetAtomMapNum()
        }

        map_to_index_numbers = {v: k for k, v in index_to_map_numbers.items()}

        # add required bonds
        for bond in rdproduct.GetBonds():
            i = map_to_index_numbers[bond.GetBeginAtom().GetAtomMapNum()]
            j = map_to_index_numbers[bond.GetEndAtom().GetAtomMapNum()]
            rwbond = rwproduct.GetBondBetweenAtoms(i, j)
            if rwbond is None:
                rwproduct.AddBond(i, j, bond.GetBondType())
            else:
                rwproduct.ReplaceBond(rwbond.GetIdx(), bond)

        for atom in rwproduct.GetAtoms():
            if atom.HasProp(ORIGINAL_MAP_NUMBER):
                atom.SetAtomMapNum(atom.GetIntProp(ORIGINAL_MAP_NUMBER))
                atom.ClearProp(ORIGINAL_MAP_NUMBER)

        rwproduct.UpdatePropertyCache()
        rwproducts.append(rwproduct)
    return rwproducts


def CleanReaction(
    reaction: rdChemReactions.ChemicalReaction,
    reactants: Tuple[rdChem.Mol, ...],
) -> Tuple[rdChemReactions.ChemicalReaction, Tuple[rdChem.Mol, ...]]:
    """Clean RDKit reaction and reactants by converting them to QueryAtoms or Atoms respectively"""
    from .rdmolops import MolAsMolWithAtoms, MolAsMolWithQueryAtoms

    new_reaction = rdChemReactions.ChemicalReaction()
    for reactant in reaction.GetReactants():
        new_reactant = MolAsMolWithQueryAtoms(reactant)
        new_reactant.UpdatePropertyCache()
        new_reaction.AddReactantTemplate(new_reactant)
    for product in reaction.GetProducts():
        new_product = MolAsMolWithQueryAtoms(product)
        new_product.UpdatePropertyCache()
        new_reaction.AddProductTemplate(new_product)

    reactants = tuple(MolAsMolWithAtoms(mol) for mol in reactants)
    for reactant in reactants:
        reactant.UpdatePropertyCache()
    return new_reaction, reactants


def RunOrClickReaction(
    reaction: rdChemReactions.ChemicalReaction,
    reactants: Tuple[rdChem.Mol, ...],
    clean: bool = True,
) -> Tuple[Tuple[rdChem.Mol, ...]]:
    """Try running reaction, or default to ClickReaction if it does not"""
    if clean:
        reaction, reactants = CleanReaction(reaction, reactants)
    products = reaction.RunReactants(reactants)
    if not products:
        products = ClickReaction(reaction, reactants)
        products = tuple((x,) for x in products)
    return products



import numpy as np
from numpy.testing import assert_allclose

from rdkit_utilities import rdMolAlign


def test_GetBestConformerRMS(methane):

    aa_rms = rdMolAlign.GetBestConformerRMS(methane, heavyAtomsOnly=False)
    assert aa_rms.shape == (5, 5)
    first_row = [0.0, 4.38178, 5.25357, 6.09918, 6.928203]
    assert_allclose(aa_rms[0], first_row, atol=1e-6)
    assert_allclose(aa_rms[:, 0], first_row, atol=1e-6)

    cg_rms = rdMolAlign.GetBestConformerRMS(methane, heavyAtomsOnly=True)
    assert cg_rms.shape == (5, 5)
    assert_allclose(cg_rms, 0, atol=1e-6)



import itertools
from typing import Optional, List

from rdkit import Chem as rdChem
import numpy as np


def GetBestConformerRMS(
    mol: rdChem.Mol, heavyAtomsOnly: bool = False, confIds: Optional[List[int]] = None
) -> np.ndarray:
    """Get square array of conformer-to-conformer symmetric RMS

    Parameters
    ----------
    mol: rdkit.Chem.Mol
        Molecule of interest
    heavyAtomsOnly: bool
        If True, all hydrogens are removed before calculating RMS.
        This is done on a copy to avoid changing the molecule
    confIds: Optional[List[int]]
        If given, the RMS calculation is limited to these conformer IDs.
        If not, all conformers are used.

    Returns
    -------
    rms: numpy.ndarray
        Square array of RMS from each conformer to each other.
        This is ordered by the input `confIds` if given, or
        by the order of the conformers on the molecule if `confIds=None`.
    """
    from rdkit.Chem import rdMolAlign

    if confIds is None:
        confIds = [c.GetId() for c in mol.GetConformers()]
    n_conformers = len(confIds)

    if heavyAtomsOnly:
        mol = rdChem.RemoveAllHs(mol, sanitize=False)
        rdChem.SanitizeMol(
            mol,
            rdChem.SANITIZE_ALL
            ^ rdChem.SANITIZE_SETAROMATICITY
            ^ rdChem.SANITIZE_SETCONJUGATION
            ^ rdChem.SANITIZE_SETHYBRIDIZATION
            ^ rdChem.SANITIZE_ADJUSTHS,
        )

    rms = np.zeros((n_conformers, n_conformers))
    for i, j in itertools.combinations(np.arange(n_conformers), 2):
        rmsd = rdMolAlign.GetBestRMS(mol, mol, confIds[i], confIds[j])
        rms[i, j] = rms[j, i] = rmsd
    return rms



"""
rdkit_utilities.py
RDKit utilities

"""
from typing import Optional, Dict
from typing_extensions import Literal

from rdkit import Chem as rdChem

from ._io._rdmolfiles import *
from ._io._parser import ALL_RDKIT_PARSERS, molecule_from_input


@reorder_constructed_molecule
def MolFromInput(
    molInput: str,
    *args,
    inputFormat: Optional[Literal[(*ALL_RDKIT_PARSERS,)]] = None,  # type: ignore
    orderByMapNumber: bool = False,
    clearAtomMapNumbers: bool = False,
    **kwargs,
) -> Optional[rdChem.Mol]:
    """Construct RDKit Mol from input with optional format.

    This accepts any input that RDKit can parse, and looks for the
    appropriate function to load the data.

    Parameters
    ----------
    molInput: str
        Molecule data or filename
    *args
        Passed to the RDKit parsing function
    inputFormat
        Format of `molInput`. If not given, it's guessed from
        the extension of the filename, or all of RDKit's content
        parsers are tried one-by-one.
    orderByMapNumber: bool
        Whether to reorder the molecule by atom map number
    clearAtomMapNumbers: bool
        Whether to remove / set all atom map numbers to 0
    """
    return molecule_from_input(molInput, *args, mol_format=inputFormat, **kwargs)



"""
rdkit_utilities
RDKit utilities
"""

try:
    from rdkit import Chem as _
except ImportError:
    raise ImportError(
        "rdkit is an required for the functionality in this package. "
        "Please install it with `conda install -c conda-forge rdkit`."
    )

from .rdchem import *
from .rdDistGeom import *
from .rdForceFieldHelpers import *
from .rdMolAlign import *
from .rdmolfiles import *
from .rdmolops import *

# Handle versioneer
from ._version import get_versions
versions = get_versions()
__version__ = versions['version']
__git_revision__ = versions['full-revisionid']
del get_versions, versions



from rdkit.Chem import rdChemReactions
from rdkit import Chem as rdChem
from rdkit_utilities import Chem
from rdkit_utilities.rdChemReactions import ClickReaction, CleanReaction


def test_run_click():
    rxn = rdChemReactions.ReactionFromSmarts(
        "[C:1](=[O:2])O.[N:3][H]>>[C:1](=[O:2])[N:3]"
    )
    cooh = "[H:5]-[C:1](=[O:2])-[O:3]-[H:4]"
    cnc = "[C:1](-[H:2])(-[H:3])(-[H:4])-[N:5](-[H:6])-[C:7](-[H:8])(-[H:9])(-[H:10])"
    reacts = (
        Chem.MolFromSmarts(
            cooh, orderByMapNumber=True, clearAtomMapNumbers=True, asQueryAtoms=False
        ),
        Chem.MolFromSmarts(
            cnc, orderByMapNumber=True, clearAtomMapNumbers=True, asQueryAtoms=False
        ),
    )
    products = ClickReaction(rxn, reacts)
    assert len(products) == 1

    product = products[0]
    expected = rdChem.AddHs(rdChem.MolFromSmiles("CN(C)C=O"))
    assert product.GetSubstructMatch(expected)
    assert expected.GetSubstructMatch(product)


def test_CleanReaction():
    rxn = rdChemReactions.ChemicalReaction()
    rct1 = rdChem.MolFromSmiles("[C:1](=[O:2])O")
    rct2 = Chem.MolFromSmiles("[N:3][H]", removeHs=False)
    prod = Chem.MolFromSmiles("[C:1](=[O:2])[N:3]")
    rxn.AddReactantTemplate(rct1)
    rxn.AddReactantTemplate(rct2)
    rxn.AddProductTemplate(prod)

    reactant1 = Chem.MolFromSmarts("[H:5]-[C:1](=[O:2])-[O:3]-[H:4]")
    reactant2 = Chem.MolFromSmarts(
        "[C:1](-[H:2])(-[H:3])(-[H:4])-[N:5](-[H:6])-[C:7](-[H:8])(-[H:9])(-[H:10])"
    )

    reaction, reactants = CleanReaction(rxn, (reactant1, reactant2))
    for reactant in reaction.GetReactants():
        for atom in reactant.GetAtoms():
            assert isinstance(atom, rdChem.QueryAtom)

    for reactant in reactants:
        for atom in reactant.GetAtoms():
            assert not isinstance(atom, rdChem.QueryAtom)

    products = reaction.RunReactants(reactants)
    assert len(products) == 1
    product = products[0][0]
    expected = rdChem.AddHs(rdChem.MolFromSmiles("CN(C)C=O"))
    assert product.GetSubstructMatch(expected)
    assert expected.GetSubstructMatch(product)



import pytest

import numpy as np
from numpy.testing import assert_allclose


def test_compute_distance_matrix():
    from rdkit_utilities import utilities

    arr = np.arange(24).reshape((2, 4, 3))
    rms = utilities.compute_atom_distance_matrix(arr)
    assert rms.shape == (2, 4, 4)

    first_row = [0, 5.19615242, 10.39230485, 15.58845727]
    assert_allclose(rms[0, 0], first_row)
    assert_allclose(rms[-1, 0], first_row)

    last_row = [15.58845727, 10.39230485, 5.19615242, 0]
    assert_allclose(rms[0, -1], last_row)
    assert_allclose(rms[-1, -1], last_row)


@pytest.mark.parametrize(
    "threshold, n_indices, indices",
    [
        (0.05, 4, [0, 3, 1, 2]),
        (0.05, 10, [0, 3, 1, 2]),
        (0.05, 2, [0, 3]),
        (10, 4, [0, 3]),
        (5.4, 4, [0, 3, 1]),
    ],
)
def test_get_maximally_diverse_indices(threshold, n_indices, indices):
    from rdkit_utilities import utilities

    arr = np.array(
        [
            [0.0, 5.49615242, 10.39230485, 15.58845727],
            [5.49615242, 0.0, 5.19615242, 10.39230485],
            [10.39230485, 5.19615242, 0.0, 5.19615242],
            [15.58845727, 10.39230485, 5.19615242, 0.0],
        ]
    )
    computed = utilities.get_maximally_diverse_indices(
        arr,
        distance_threshold=threshold,
        n_indices=n_indices,
    )
    assert computed == indices



import pytest

import numpy as np
from numpy.testing import assert_allclose
from rdkit import Chem as rdChem

from rdkit_utilities import rdmolfiles, rdmolops


def test_OrderByMapNumber():
    mol = rdmolfiles.MolFromSmiles("[C:3][C:2][O:1]")
    assert mol.GetAtomWithIdx(0).GetSymbol() == "C"

    reordered = rdmolops.OrderByMapNumber(mol, clearAtomMapNumbers=False)
    first = reordered.GetAtomWithIdx(0)
    assert first.GetSymbol() == "O"
    assert first.GetAtomMapNum() == 1

    reordered = rdmolops.OrderByMapNumber(mol, clearAtomMapNumbers=True)
    first = reordered.GetAtomWithIdx(0)
    assert first.GetSymbol() == "O"
    assert first.GetAtomMapNum() == 0


@pytest.mark.parametrize(
    "core_indices, include_central_atoms, n_neighbors, expected_indices",
    [
        ({14}, True, -1, set(range(25))),
        ({14}, False, -1, set(range(14)) | set(range(15, 25))),
        ({14}, True, 0, {14}),
        ({14}, False, 0, set()),
        ({14}, True, 1, {12, 14, 15}),
        ({14}, False, 1, {12, 15}),
        ({14}, True, 2, {7, 13, 12, 14, 15, 16, 17, 18}),
        ({14}, False, 2, {7, 13, 12, 15, 16, 17, 18}),
        ({14}, True, 3, {5, 8, 7, 13, 12, 14, 15, 16, 17, 18, 19, 20, 21}),
        ({14}, False, 3, {5, 8, 7, 13, 12, 15, 16, 17, 18, 19, 20, 21}),
        ({2, 12, 21}, True, -1, set(range(25))),
        ({2, 12, 21}, True, 0, {2, 12, 21}),
        ({2, 12, 21}, False, 0, set()),
        ({2, 12, 21}, False, 1, {1, 3, 10, 7, 13, 14, 18, 22, 23, 24}),
        ({2, 12}, False, 2, {0, 1, 3, 4, 5, 10, 8, 11, 7, 13, 14, 5, 8, 15}),
    ],
)
def test_GetAtomNeighborIndices(
    propylparaben, core_indices, include_central_atoms, n_neighbors, expected_indices
):
    indices = rdmolops.GetAtomNeighborIndices(
        propylparaben,
        centralAtomIndices=core_indices,
        includeCentralAtoms=include_central_atoms,
        numAtomNeighbors=n_neighbors,
    )
    assert indices == expected_indices


@pytest.fixture()
def mol_with_conformers():
    from rdkit_utilities.rdchem import AddConformerWithCoordinates

    mol = rdmolfiles.MolFromSmarts("[O]=[C]=[O]")
    coords = np.zeros((3, 3), dtype=float)
    for i in range(10):
        AddConformerWithCoordinates(mol, coords + i)
    return mol


def test_ReorderConformers(mol_with_conformers):
    conf = mol_with_conformers.GetConformer(0)
    assert_allclose(np.array(conf.GetPositions())[0], [0, 0, 0])
    assert conf.GetId() == 0

    order = [4, 2, 1, 0, 3, 5, 6, 7, 8, 9]
    rdmolops.ReorderConformers(mol_with_conformers, order)
    conf = mol_with_conformers.GetConformer(0)
    assert_allclose(np.array(conf.GetPositions())[0], [4, 4, 4])
    assert conf.GetId() == 0


def test_KeepConformerIds(mol_with_conformers):
    rdmolops.KeepConformerIds(mol_with_conformers, [1, 0, 3, 5, 6, 7])
    conformer_ids = [conf.GetId() for conf in mol_with_conformers.GetConformers()]
    assert conformer_ids == [0, 1, 3, 5, 6, 7]


@pytest.mark.parametrize(
    "atom_indices_in, atom_nums_out",
    [([], []), ([0, 1], [1, 2]), ([5, 0, 3, 2], [1, 3, 4]), ((3, 4, 4), [4, 5])],
)
def test_SubsetMol(methane, atom_indices_in, atom_nums_out):
    subset = rdmolops.SubsetMol(methane, atom_indices_in)
    assert methane.GetNumAtoms() == 5

    subset_nums = [atom.GetAtomMapNum() for atom in subset.GetAtoms()]
    assert subset_nums == atom_nums_out


def test_MolToMolWithAtoms():
    query = rdmolfiles.MolFromSmarts("[*&r5]-[c:3]")
    assert all(isinstance(a, rdChem.QueryAtom) for a in query.GetAtoms())
    for atom in query.GetAtoms():
        atom.SetIsotope(10)
    last = query.GetAtomWithIdx(1)
    last.SetAtomMapNum(5)
    last.SetProp("dummy", "foo")

    mol = rdmolops.MolToMolWithAtoms(query)
    assert all(isinstance(a, rdChem.Atom) for a in mol.GetAtoms())

    assert mol.GetNumAtoms() == 2
    at1, at2 = list(mol.GetAtoms())
    q1, q2 = list(query.GetAtoms())

    assert q1.GetAtomicNum() == 0
    assert q1.GetSymbol() == "*"
    assert not q1.IsInRing()  # shouldn't show up in mol
    assert not q1.GetIsAromatic()
    assert not q1.GetAtomMapNum()
    assert q1.GetIsotope() == 10
    assert q1.GetSmarts() == "[r5]"

    assert at1.GetAtomicNum() == 0
    assert at1.GetSymbol() == "*"
    assert not at1.IsInRing()  # shouldn't show up in mol
    assert not at1.GetIsAromatic()
    assert not at1.GetAtomMapNum()
    assert at1.GetIsotope() == 10

    assert q2.GetAtomicNum() == 6
    assert q2.GetSymbol() == "C"
    assert q2.GetIsAromatic()
    assert q2.GetAtomMapNum() == 5
    assert q2.GetIsotope() == 10

    assert at2.GetAtomicNum() == 6
    assert at2.GetSymbol() == "C"
    assert at2.GetIsAromatic()
    assert at2.GetAtomMapNum() == 5
    assert at2.GetIsotope() == 10
    assert at2.GetProp("dummy") == "foo"

    assert rdChem.MolToSmarts(query) == "[r5]-[c:5]"
    assert rdChem.MolToSmarts(mol) == "[10#0]-[10#6:5]"

    return_trip = rdmolops.MolToMolWithQueryAtoms(mol)
    assert rdChem.MolToSmarts(return_trip) == "[r5&10*]-[c&10*:5]"


@pytest.mark.parametrize(
    "strict, includeIsotopes, smarts",
    [
        # Atom.GetSmarts contains isotopes even though QueryAtom.GetSmarts does not
        (False, False, "[C&H1](=[O&4*])-[O&-]"),
        (False, True, "[C&H1](=[O&4*])-[O&-]"),
        (True, True, "[C&H1&0*](=[O&4*])-[O&-&0*]"),
    ],
)
def test_MolToMolWithQueryAtoms_from_smiles(strict, includeIsotopes, smarts):
    mol = rdChem.MolFromSmiles("C(=[4O])-[O-]")
    assert mol.GetNumAtoms() == 3
    at1, at2, at3 = list(mol.GetAtoms())
    assert at2.GetIsotope() == 4
    assert at2.GetSmarts() == "[4O]"
    assert at3.GetIsotope() == 0
    at1.SetProp("dummy", "foo")

    query = rdmolops.MolToMolWithQueryAtoms(
        mol, strict=strict, includeIsotopes=includeIsotopes
    )
    assert query.GetAtomWithIdx(0).GetProp("dummy") == "foo"
    output_smarts = rdChem.MolToSmarts(query)
    assert output_smarts == smarts


@pytest.mark.parametrize(
    "strict, includeIsotopes, smarts",
    [
        # Atom.GetSmarts contains isotopes even though QueryAtom.GetSmarts does not
        (False, False, "C(=O)-[O&-:5]"),
        (False, True, "C(=[O&4*])-[O&-:5]"),
        (True, True, "[C&0*](=[O&4*])-[O&-&0*:5]"),
    ],
)
def test_MolToMolWithQueryAtoms_from_smarts(strict, includeIsotopes, smarts):
    mol = rdChem.MolFromSmarts("C(=[O])-[O-]")
    assert mol.GetNumAtoms() == 3
    at1, at2, at3 = list(mol.GetAtoms())
    assert at1.GetSmarts() == "C"
    at2.SetIsotope(4)
    assert at2.GetSmarts() == "O"
    at3.SetAtomMapNum(5)
    assert at3.GetSmarts() == "[O&-:5]"

    query = rdmolops.MolToMolWithQueryAtoms(
        mol, strict=strict, includeIsotopes=includeIsotopes
    )
    output_smarts = rdChem.MolToSmarts(query)
    assert output_smarts == smarts

    normal_mol = rdmolops.MolToMolWithAtoms(mol)
    query_from_normal = rdmolops.MolToMolWithQueryAtoms(
        normal_mol, strict=strict, includeIsotopes=includeIsotopes
    )
    output_smarts_from_normal = rdChem.MolToSmarts(query_from_normal)
    assert output_smarts_from_normal == smarts


def test_MolToMolIdentity():
    smiles = rdChem.MolFromSmiles("CC")
    smiles_to_smiles = rdmolops.MolAsMolWithAtoms(smiles)
    assert smiles_to_smiles is smiles

    smiles_to_smarts = rdmolops.MolAsMolWithQueryAtoms(smiles)
    assert smiles_to_smarts is not smiles
    assert rdChem.MolToSmarts(smiles_to_smarts) == "[C&H3]-[C&H3]"

    smarts = rdChem.MolFromSmarts("CC")
    smarts_to_smiles = rdmolops.MolAsMolWithAtoms(smarts)
    assert smarts_to_smiles is not smarts
    assert rdChem.MolToSmarts(smarts) == "CC"

    smarts_to_smarts = rdmolops.MolAsMolWithQueryAtoms(smarts)
    assert smarts_to_smarts is smarts



from typing import List, Set, Union

import numpy as np
from rdkit import Chem as rdChem

_RDPROP_ATOM_SMARTS = "_smarts_query"


def GetAtomNeighborIndices(
    mol: rdChem.Mol,
    centralAtomIndices: Union[Set[int], List[int]] = [],
    includeCentralAtoms: bool = True,
    numAtomNeighbors: int = 0,
) -> Set[int]:
    """
    Get neighbor atom indices around specified core

    Parameters
    ----------
    mol: rdkit.Chem.Mol
    centralAtomIndices: List[int]
        Central atoms to get neighbors around
    includeCentralAtoms: bool
        Whether to include the central atoms in the output
    numAtomNeighbors: int
        Size of shell around the central core.
        -1 returns all indices in the molecule;
        0 returns no additional neighbors;
        1 returns the first shell around the central core,
        and so on.

    Returns
    -------
    Set[int]
        Set of neighbor atom indices around specified core
    """
    central_atoms = set(centralAtomIndices)
    neighbor_atoms = set(central_atoms)

    if numAtomNeighbors < 0:
        neighbor_atoms = set(range(mol.GetNumAtoms()))

    else:
        current_layer = central_atoms
        while numAtomNeighbors:
            new_layer = set()
            for index in current_layer:
                atom = mol.GetAtomWithIdx(index)
                for bond in atom.GetBonds():
                    new_layer.add(bond.GetOtherAtomIdx(index))

            new_layer -= neighbor_atoms
            neighbor_atoms |= new_layer
            current_layer = new_layer
            numAtomNeighbors -= 1

    if not includeCentralAtoms:
        neighbor_atoms -= central_atoms
    return neighbor_atoms


def OrderByMapNumber(
    mol: rdChem.Mol,
    clearAtomMapNumbers: bool = True,
) -> rdChem.Mol:
    """
    Reorder RDKit molecule by atom map number. This returns a copy.

    Parameters
    ----------
    mol: rdkit.Chem.Mol
        RDKit molecule
    clearAtomMapNumbers: bool
        Whether to set atom map numbers in the output
        molecule to 0

    Returns
    -------
    rdkit.Chem.Mol
        Output molecule. This is a **copy** of the input molecule.

    Example
    -------

    ::

        >>> rdmol = Chem.MolFromSmiles("[C:3][C:2][O:1]")
        >>> rdmol.GetAtomWithIdx(0).GetSymbol()
        'C'
        >>> reordered = OrderByMapNumber(rdmol)
        >>> reordered.GetAtomWithIdx(0).GetSymbol()
        'O'
    """
    map_numbers = [atom.GetAtomMapNum() for atom in mol.GetAtoms()]
    order = list(map(int, np.argsort(map_numbers)))
    reordered = rdChem.RenumberAtoms(mol, order)
    if clearAtomMapNumbers:
        for atom in reordered.GetAtoms():
            atom.SetAtomMapNum(0)
    return reordered


def ReorderConformers(
    mol: rdChem.Mol, order: Union[List[int], np.ndarray], resetConfId: bool = True
):
    """Reorder conformers in-place by `order`

    Parameters
    ----------
    mol: rdkit.Chem.Mol
        Molecule of interest
    order: Union[List[int], np.ndarray]
        New order. This should be indexed from 0,
        and is typically the output of `np.argsort`.
        If an index is not present in this list,
        the associated conformer is removed from the molecule.
    resetConfId: bool
        Whether to reset the conformer ID so they are sequential.
    """
    conformers = [rdChem.Conformer(x) for x in mol.GetConformers()]
    mol.RemoveAllConformers()
    for new_index, current_index in enumerate(order):
        conformer = conformers[current_index]
        if resetConfId:
            conformer.SetId(int(new_index))
        mol.AddConformer(conformer)


def KeepConformerIds(mol: rdChem.Mol, confIds: Union[List[int], np.ndarray]):
    """Remove conformers from `mol` except those in `confIds`"""
    conf_ids = [conf.GetId() for conf in mol.GetConformers()]
    to_remove = [x for x in conf_ids if x not in confIds]
    for confid in to_remove[::-1]:
        mol.RemoveConformer(confid)


def SubsetMol(
    mol: rdChem.Mol,
    atomIndices: List[int],
    reorder: bool = False,
    clear_atom_valences: bool = True,
) -> rdChem.Mol:
    """Return a subset of a molecule, as a copy"""
    mol = rdChem.RWMol(mol)
    all_indices = list(range(mol.GetNumAtoms()))
    to_keep = [i for i in atomIndices if i in all_indices]
    to_delete = [i for i in all_indices if i not in to_keep]
    if reorder:
        new_order = tuple(to_keep + to_delete)
        mol = rdChem.RWMol(rdChem.RenumberAtoms(mol, new_order))
        to_delete = range(len(to_keep), mol.GetNumAtoms())

    for index in to_delete[::-1]:
        mol.RemoveAtom(index)

    if clear_atom_valences:
        for atom in mol.GetAtoms():
            atom.SetNumRadicalElectrons(0)
            atom.SetNoImplicit(False)
            atom.UpdatePropertyCache()

    mol.UpdatePropertyCache()
    return rdChem.Mol(mol)


def AtomFromQueryAtom(atom: rdChem.QueryAtom) -> rdChem.Atom:
    """Convert Chem.QueryAtom to Chem.Atom"""
    # this is a separate function just in case I want to add in future...
    from .rdchem import SetPropsFromDict
    new_atom = rdChem.Atom(atom)
    new_atom.SetProp(_RDPROP_ATOM_SMARTS, atom.GetSmarts())
    SetPropsFromDict(new_atom, atom.GetPropsAsDict())
    return new_atom


def AtomToQueryAtom(
    atom: rdChem.Atom,
    strict: bool = False,
    includeIsotope: bool = True,
) -> rdChem.QueryAtom:
    """Convert Chem.Atom to Chem.QueryAtom"""
    from rdkit.Chem import rdqueries
    from .rdchem import SetPropsFromDict

    if atom.HasProp(_RDPROP_ATOM_SMARTS):
        q = rdChem.AtomFromSmarts(atom.GetProp(_RDPROP_ATOM_SMARTS))
    else:
        # this is option #2 because it can infer implicit Hs
        # that were not originally there
        atom.UpdatePropertyCache(strict=True)
        # allHsExplicit is for when it dips to MolToSmiles when Smarts fails
        q = rdChem.AtomFromSmarts(atom.GetSmarts(allHsExplicit=True))

    q.SetAtomMapNum(atom.GetAtomMapNum())

    # TODO: investigate including other properties
    # like H count and stuff
    if includeIsotope:
        iso = atom.GetIsotope()
        if strict or iso:
            qiso = rdqueries.IsotopeEqualsQueryAtom(iso)
            pattern = qiso.GetSmarts().strip("[]")
            if pattern not in q.GetSmarts():
                q.ExpandQuery(qiso)

    props = atom.GetPropsAsDict()
    SetPropsFromDict(q, props)
    return q


def MolToMolWithAtoms(mol: rdChem.Mol) -> rdChem.Mol:
    """Convert Mol with QueryAtoms to Mol with Atoms"""
    normal_mol = rdChem.RWMol(mol)
    for i, atom in enumerate(mol.GetAtoms()):
        normal_mol.ReplaceAtom(i, AtomFromQueryAtom(atom))
    # normal_mol.ClearComputedProperties()
    return rdChem.Mol(normal_mol)


def MolToMolWithQueryAtoms(
    mol: rdChem.Mol, strict: bool = False, includeIsotopes: bool = True
) -> rdChem.Mol:
    """Convert Mol with Atoms to Mol with QueryAtoms"""
    query_mol = rdChem.RWMol(mol)
    for i, atom in enumerate(mol.GetAtoms()):
        q = AtomToQueryAtom(atom, strict=strict, includeIsotope=includeIsotopes)
        query_mol.ReplaceAtom(i, q)
    # query_mol.ClearComputedProperties()
    return rdChem.Mol(query_mol)


def MolAsMolWithAtoms(mol: rdChem.Mol) -> rdChem.Mol:
    """If Mol contains Chem.QueryAtoms, convert to Mol with Chem.Atoms

    If Mol already is composed of Atoms, returns the same object
    """
    if not any(isinstance(atom, rdChem.QueryAtom) for atom in mol.GetAtoms()):
        return mol
    return MolToMolWithAtoms(mol)


def MolAsMolWithQueryAtoms(
    mol: rdChem.Mol, strict: bool = False, includeIsotopes: bool = True
) -> rdChem.Mol:
    """If Mol contains Chem.Atoms, convert to Mol with Chem.QueryAtoms

    If Mol already is composed of QueryAtoms, returns the same object
    """
    if all(isinstance(atom, rdChem.QueryAtom) for atom in mol.GetAtoms()):
        return mol
    return MolToMolWithQueryAtoms(mol, strict=strict, includeIsotopes=includeIsotopes)



import re
import pytest

import numpy as np
from numpy.testing import assert_allclose

from rdkit import Chem as rdChem
from rdkit_utilities import rdchem, rdmolfiles
from rdkit_utilities.tests.conftest import propylparaben


@pytest.mark.parametrize(
    "smarts, indices",
    [
        ("[C](=[O])-[O-]", [(1, 2)]),
        (
            "[N:1](-[H:2])(-[H:3])-[C:4](-[O:5]-[H:6])(-[O:7]-[H:8])-[O:9]-[H:10]",
            [(1, 2), (4, 6, 8), (5, 7, 9)],
        ),
        (
            "[N-2:1]-[C:2](-[O:3]-[H:4])(-[O:5]-[H:6])-[O:7]-[H:8]",
            [(2, 4, 6), (3, 5, 7)],
        ),
    ],
)
def test_GetSymmetricAtomIndices(smarts, indices):
    mol = rdmolfiles.MolFromSmarts(
        smarts, orderByMapNumber=True, clearAtomMapNumbers=True
    )
    rdChem.SanitizeMol(mol)
    assert rdchem.GetSymmetricAtomIndices(mol) == indices


def test_AddConformerWithCoordinates(propylparaben):
    assert propylparaben.GetNumConformers() == 0
    coordinates = np.ones((25, 3))
    rdchem.AddConformerWithCoordinates(propylparaben, coordinates)
    assert propylparaben.GetNumConformers() == 1
    rdcoords = propylparaben.GetConformer(0).GetPositions()
    assert_allclose(np.array(rdcoords), coordinates)


def test_AddConformerWithCoordinates_error(propylparaben):
    assert propylparaben.GetNumConformers() == 0
    coordinates = np.ones((24, 3))

    err = "Shape of coordinates must be (25, 3). " "Given array with shape (24, 3)"

    with pytest.raises(ValueError, match=re.escape(err)):
        rdchem.AddConformerWithCoordinates(propylparaben, coordinates)
    assert propylparaben.GetNumConformers() == 0


@pytest.mark.parametrize(
    "smarts, indices, mapping",
    [
        ("[H:1]-[O:2]", [(0, 1)], [{1: 0, 2: 1}]),
        ("[H:4]-[O:2]-[c]", [(1, 0)], [{2: 1, 4: 0}]),
        (
            "[C:2](-[H:4])(-[H:6])-[H]",
            [
                (21, 22, 23),
                (21, 23, 22),
                (21, 22, 24),
                (21, 24, 22),
                (21, 23, 24),
                (21, 24, 23),
            ],
            [
                {2: 21, 4: 22, 6: 23},
                {2: 21, 4: 23, 6: 22},
                {2: 21, 4: 22, 6: 24},
                {2: 21, 4: 24, 6: 22},
                {2: 21, 4: 23, 6: 24},
                {2: 21, 4: 24, 6: 23},
            ],
        ),
    ],
)
def test_GetTaggedSubstructMatches(propylparaben, smarts, indices, mapping):
    query = rdChem.MolFromSmarts(smarts)
    as_list = rdchem.GetTaggedSubstructMatches(propylparaben, query)
    assert set(as_list) == set(indices)

    as_dict = rdchem.GetTaggedSubstructMatches(
        propylparaben, query, mappingAsTaggedDict=True
    )
    assert len(as_dict) == len(mapping)
    for mapping_ in mapping:
        assert mapping_ in as_dict


def test_SetPropsFromDict():
    properties = dict(
        boolean_prop1=True,
        boolean_prop2=False,
        integer_prop1=0,
        integer_prop2=1,
        integer_prop3=-10,
        np_int_prop=np.int64(4),
        float_prop=3.0,
        np_float_prop=np.float64(12),
        str_prop="asdf",
        empty_str_prop="",
    )

    atom = rdChem.Atom("H")
    rdchem.SetPropsFromDict(atom, properties)
    atom_properties = atom.GetPropsAsDict()
    assert properties == atom_properties

    # manually go through this because in Windows, !np.issubdtype(int64, int)
    assert type(atom_properties["boolean_prop1"]) is bool
    assert type(atom_properties["boolean_prop2"]) is bool
    assert type(atom_properties["integer_prop1"]) is int
    assert type(atom_properties["integer_prop2"]) is int
    assert type(atom_properties["integer_prop3"]) is int
    assert type(atom_properties["np_int_prop"]) is int
    assert type(atom_properties["float_prop"]) is float
    assert type(atom_properties["np_float_prop"]) is float
    assert type(atom_properties["str_prop"]) is str
    assert type(atom_properties["empty_str_prop"]) is str

    mol = rdChem.MolFromSmiles("CC")
    rdchem.SetPropsFromDict(mol, properties)
    mol_properties = mol.GetPropsAsDict()
    assert properties == mol_properties
    
    assert type(mol_properties["boolean_prop1"]) is bool
    assert type(mol_properties["boolean_prop2"]) is bool
    assert type(mol_properties["integer_prop1"]) is int
    assert type(mol_properties["integer_prop2"]) is int
    assert type(mol_properties["integer_prop3"]) is int
    assert type(mol_properties["np_int_prop"]) is int
    assert type(mol_properties["float_prop"]) is float
    assert type(mol_properties["np_float_prop"]) is float
    assert type(mol_properties["str_prop"]) is str
    assert type(mol_properties["empty_str_prop"]) is str


import pytest

import numpy as np
from numpy.testing import assert_allclose

from rdkit_utilities import rdmolfiles, rdchem, rdDistGeom
from rdkit import Chem as rdChem


@pytest.fixture
def formic_acid():
    mol = rdmolfiles.MolFromSmiles(
        "[H:4][C:2](=[O:1])[O:3][H:5]",
        orderByMapNumber=True,
        removeHs=False,
        clearAtomMapNumbers=True,
    )
    return mol


def test_SelectDiverseELFConformers():
    smiles = "[H:6]/[C:1](=[C:2](\\[H:7])/[O:3][H:8])/[C:4](=[O:5])[H:9]"
    mol = rdmolfiles.MolFromSmiles(smiles, orderByMapNumber=True, removeHs=False)
    conformers = np.array(
        [
            [
                [0.5477, 0.3297, -0.0621],
                [-0.1168, -0.7881, 0.2329],
                [-1.4803, -0.8771, 0.1667],
                [-0.2158, 1.5206, -0.4772],
                [-1.4382, 1.5111, -0.5580],
                [1.6274, 0.3962, -0.0089],
                [0.3388, -1.7170, 0.5467],
                [-1.8612, -0.0347, -0.1160],
                [0.3747, 2.4222, -0.7115],
            ],
            [
                [0.5477, 0.3297, -0.0621],
                [-0.1168, -0.7881, 0.2329],
                [-1.4803, -0.8771, 0.1667],
                [-0.2158, 1.5206, -0.4772],
                [0.3353, 2.5772, -0.7614],
                [1.6274, 0.3962, -0.0089],
                [0.3388, -1.7170, 0.5467],
                [-1.7743, -1.7634, 0.4166],
                [-1.3122, 1.4082, -0.5180],
            ],
        ]
    )
    for conf in conformers:
        rdchem.AddConformerWithCoordinates(mol, conf)
    assert mol.GetNumConformers() == 2

    rdDistGeom.SelectDiverseELFConformers(mol)
    assert mol.GetNumConformers() == 1

    selected = np.array(mol.GetConformer(0).GetPositions())
    assert_allclose(selected, conformers[1])


def test_RemoveTransAcidConformers(formic_acid):
    cis = np.array(
        [
            [-0.95927322, -0.91789997, 0.36333418],
            [-0.34727824, 0.12828046, 0.22784603],
            [0.82766682, 0.26871252, -0.42284882],
            [-0.67153811, 1.10376000, 0.61921501],
            [1.15035689, -0.58282924, -0.78766006],
        ]
    )
    trans = np.array(
        [
            [-0.95927322, -0.91789997, 0.36333418],
            [-0.34727824, 0.12828046, 0.22784603],
            [0.82766682, 0.26871252, -0.42284882],
            [-0.67153811, 1.10376000, 0.61921501],
            [1.14532626, 1.19679034, -0.41266876],
        ]
    )
    rdchem.AddConformerWithCoordinates(formic_acid, cis)
    rdchem.AddConformerWithCoordinates(formic_acid, trans)

    assert formic_acid.GetNumConformers() == 2
    rdDistGeom.RemoveTransAcidConformers(formic_acid)
    assert formic_acid.GetNumConformers() == 1

    coordinates = np.array(formic_acid.GetConformer(0).GetPositions())
    assert_allclose(coordinates, cis)


def test_CalculateElectrostaticEnergy(formic_acid):
    assert formic_acid.GetNumConformers() == 0

    dummy1 = np.array(
        [
            [1.0, 0.0, 0.0],  # =O
            [0.0, 0.0, 0.0],  # C
            [-1.0, 0.0, 0.0],  # -O-
            [0.0, 1.0, 0.0],  # H-(C)
            [0.0, -1.0, 0.0],  # H-(O)
        ]
    )

    # swap O
    dummy2 = np.array(
        [
            [-1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, -1.0, 0.0],
        ]
    )

    # swap H
    dummy3 = np.array(
        [
            [-1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, -1.0, 0.0],
            [0.0, 1.0, 0.0],
        ]
    )

    # shake it up
    dummy4 = np.array(
        [
            [-1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, -1.0, 0.0],
            [1.0, 0.0, 0.0],
        ]
    )

    rdchem.AddConformerWithCoordinates(formic_acid, dummy1)
    rdchem.AddConformerWithCoordinates(formic_acid, dummy2)
    rdchem.AddConformerWithCoordinates(formic_acid, dummy3)
    rdchem.AddConformerWithCoordinates(formic_acid, dummy4)
    assert formic_acid.GetNumConformers() == 4

    energies = rdDistGeom.CalculateElectrostaticEnergy(formic_acid)
    assert energies.shape == (4,)
    desired_energies = [0.216525, 0.216525, 0.216525, 0.163713]
    assert_allclose(energies, desired_energies, atol=1e-6)


def test_GenerateConformers(formic_acid):
    assert formic_acid.GetNumConformers() == 0
    rdDistGeom.GenerateConformers(formic_acid, randomSeed=1)
    assert formic_acid.GetNumConformers() == 10

    lowest = formic_acid.GetConformer(0)
    coordinates = np.array(lowest.GetPositions())
    first_coord = [-0.677714, 1.017283, 0.609685]
    assert_allclose(coordinates[0], first_coord, atol=1e-6)


@pytest.mark.parametrize(
    "numConfPool, maximizeDiversity, selectELFConfs, optimizeConfs, n_confs",
    [
        (None, False, False, False, 10),
        (200, False, False, False, 10),
        (200, False, True, False, 4),
        (1000, False, True, False, 10),  # 10 selected from 20
        (200, True, False, False, 10),
        (10000, True, True, True, 10),
    ],
)
def test_GenerateConformers_options(
    numConfPool, maximizeDiversity, selectELFConfs, optimizeConfs, n_confs
):
    mol = rdChem.AddHs(rdmolfiles.MolFromSmiles("CCCCCC"))
    rdDistGeom.GenerateConformers(
        mol,
        numConfPool=numConfPool,
        maximizeDiversity=maximizeDiversity,
        selectELFConfs=selectELFConfs,
        optimizeConfs=optimizeConfs,
        diverseRmsThresh=0.01,
    )
    assert mol.GetNumConformers() == n_confs



import functools
from typing import Optional, Dict

from rdkit import Chem as rdChem


__all__ = [
    "reorder_constructed_molecule",
    "MolFromSmiles",
    "MolFromSmarts",
]


def reorder_constructed_molecule(func):
    @functools.wraps(func)
    def wrapper(
        *args,
        orderByMapNumber: bool = False,
        clearAtomMapNumbers: bool = False,
        **kwargs
    ):
        mol = func(*args, **kwargs)
        if orderByMapNumber and mol is not None:
            from rdkit_utilities.rdmolops import OrderByMapNumber

            mol = OrderByMapNumber(mol, clearAtomMapNumbers=clearAtomMapNumbers)
        elif clearAtomMapNumbers:
            for atom in mol.GetAtoms():
                atom.SetAtomMapNum(0)
        return mol

    return wrapper


@reorder_constructed_molecule
def MolFromSmiles(
    smiles: str,
    allowCXSMILES: bool = True,
    maxIterations: int = 0,
    parseName: bool = False,
    removeHs: bool = True,
    sanitize: bool = True,
    strictCXSMILES: bool = True,
    useLegacyStereo: bool = True,
    orderByMapNumber: bool = False,
    clearAtomMapNumbers: bool = False,
) -> Optional[rdChem.Mol]:
    """
    Create a molecule from a SMILES string.

    Parameters
    ----------
    smiles: str
    orderByMapNumber: bool
        Whether to reorder the molecule by atom map number
    clearAtomMapNumbers: bool
        Whether to remove / set all atom map numbers to 0
    **kwargs
        Passed to rdkit.Chem.SmilesParserParams

    Example
    -------

    ::

        >>> rdmol = MolFromSmiles("[C:3][C:2][H:1]")
        >>> rdmol.GetNumAtoms()
        2
        >>> rdmol = MolFromSmiles("[C:3][C:2][H:1]", removeHs=False)
        >>> rdmol.GetNumAtoms()
        3
    """
    smiles_parser = rdChem.rdmolfiles.SmilesParserParams()
    smiles_parser.allowCXSMILES = allowCXSMILES
    smiles_parser.maxIterations = maxIterations
    smiles_parser.parseName = parseName
    smiles_parser.removeHs = removeHs
    smiles_parser.sanitize = sanitize
    smiles_parser.strictCXSMILES = strictCXSMILES
    smiles_parser.useLegacyStereo = useLegacyStereo
    mol = rdChem.MolFromSmiles(smiles, smiles_parser)
    return mol


@reorder_constructed_molecule
def MolFromSmarts(
    smarts: str,
    mergeHs: bool = False,
    replacements: Dict[str, str] = {},
    asQueryAtoms: bool = True,
    orderByMapNumber: bool = False,
    clearAtomMapNumbers: bool = False,
) -> Optional[rdChem.Mol]:
    """
    Create a molecule from a SMARTS string.

    Parameters
    ----------
    smarts: str
    orderByMapNumber: bool
        Whether to reorder the molecule by atom map number
    clearAtomMapNumbers: bool
        Whether to remove / set all atom map numbers to 0
    **kwargs
        Passed to rdkit.Chem.SmilesParserParams

    """
    mol = rdChem.MolFromSmarts(smarts, mergeHs=mergeHs, replacements=replacements)
    if not asQueryAtoms and mol is not None:
        from ..rdmolops import MolToMolWithAtoms

        return MolToMolWithAtoms(mol)
    return mol
